#include <cmath>
#include <cassert>

#include <algorithm>
#include <stdexcept>

#include "constants.h"
#include "electrostatics.h"

//#define DEBUG

#ifdef DEBUG
#  include <iomanip>
#  include <iostream>
#endif // DEBUG

//const double DEBYE = constants::DEBYE;

////////////////////////////////////////////////////////////////////////////////

namespace {

// DAMPING PARAMETERS for CC, CD and DD for neutral-neutral,
// ion-ion and neutral ion interactions
double CCnn = 0.4;
double CCni = 0.4;
double CCii = 0.4;
double CDnn = 0.4;
double CDni = 0.4;
double CDii = 0.4;
double DDnn = 0.055;
double DDni = 0.055;
double DDii = 0.055;

bool is_excluded(const int a, const int b, 
                 const excluded_set_type& excluded_ab)
{
    return (    (excluded_ab.find(std::make_pair(a, b)) != excluded_ab.end())
            ||  (excluded_ab.find(std::make_pair(b, a)) != excluded_ab.end()) );
}

} // namespace

////////////////////////////////////////////////////////////////////////////////

namespace ttm {

//----------------------------------------------------------------------------//

electrostatics::electrostatics() //constructor with no arguments 
: m_size(0)
{
    allocate(0);
}

//----------------------------------------------------------------------------//

electrostatics::~electrostatics() //destructor
{
    allocate(0);
}

//----------------------------------------------------------------------------//

void electrostatics::allocate(size_t n) 
{
    if (m_size == n) {
        return;
    }

    if (m_size > 0) {
        delete[] m_memory;
    }

    m_size = n;

    if (n > 0)
        m_memory = new double[n*(6*3 + 1 + 9*n)];
}

//----------------------------------------------------------------------------//

const double* electrostatics::dipoles() const
{
    return m_memory;
}

//----------------------------------------------------------------------------//

double electrostatics::operator()(const sys::system& my_system,
                                  const smear& smr,
                                  double* grad)
{
    return electrostatics::operator()(my_system.nsites, my_system.charges,
                                      my_system.polfac, my_system.atmpolar,
                                      my_system.coords,
                                      my_system.excluded12,
                                      my_system.excluded13,
                                      my_system.first_index_of_molec,
                                      my_system.molec,
                                      smr, grad);
}

//----------------------------------------------------------------------------//
// MRR electrostatics modification to add excluded 14

double electrostatics::operator()(size_t natom, const double* charge,
    const double* polfac, const double* polar, const double* xyz,
    const excluded_set_type& excluded12,
    const excluded_set_type& excluded13,
    const excluded_set_type& excluded14,
    const std::vector<int>& first_index_of_molec,
    const std::vector<std::shared_ptr<molecule>>& molec, const smear& smr,
    double* grad)
{

    allocate(natom);

    double* dipole      = m_memory;

    double* polar_sqrt  = dipole + 3*natom;
    double* phi         = polar_sqrt + 3*natom;
    double* Efq         = phi + natom;
    double* Efd         = Efq + 3*natom;
    double* ddt         = Efd + 3*natom;

    // FIXME: This is the wrong length?
    std::fill(phi, phi + natom*(1 + 2*3 + 3*3*natom), 0.0);

    // Necessary to check if the ADD must be aDD-nn, -ni or -ii.
    size_t nmol = 0;
    size_t count = 0;
    std::vector<std::string> n_i_vector;
    for (size_t i = 0; i < natom; ++i) {
      if (count >= molec[nmol]->get_nsites()) {
        nmol++;
        count = 0;
      }
      if (molec[nmol]->id == "n") {
        n_i_vector.push_back("n");
      } else if (molec[nmol]->id == "i") {
        n_i_vector.push_back("i");
      }
      count++;
    } 

    for (size_t i = 0; i < natom; ++i) {
        const size_t i3 = 3*i;
        const double sqrtalpha = std::sqrt(polar[i]);

        for (size_t k = 0; k < 3; ++k)
            polar_sqrt[i3 + k] = sqrtalpha;
    }

    for (size_t i = 0; i < natom; ++i) {
        const size_t i3 = 3*i;

        for (size_t j = i + 1; j < natom; ++j) {
            const size_t j3 = 3*j;

            double Rij[3], Rsq(0);
            for (size_t k = 0; k < 3; ++k) {
                Rij[k] = xyz[i3 + k] - xyz[j3 + k];
                Rsq += Rij[k]*Rij[k];
            }

            // charge-charge

            const bool ij_12 = is_excluded(i, j, excluded12);
            const bool ij_13 = is_excluded(i, j, excluded13);
            const bool ij_14 = is_excluded(i, j, excluded14);

            const bool ij_bonded = ij_12 || ij_13 || ij_14;

            if (!ij_bonded) {
                double ts0, ts1;
                double aCC =  smr.aCC();
                if ( !ij_bonded) {
                  if (n_i_vector[i] == "n" && n_i_vector[j] == "n") {
                    aCC = CCnn;
                  } else if (n_i_vector[i] == "n" || n_i_vector[j] == "n"){
                    aCC = CCni;
                  } else if (n_i_vector[i] == "i" || n_i_vector[j] == "i"){
                    aCC = CCii;
                  } else {
                    throw std::logic_error("molecule type not defined");
                  }
                }

                smr.smear01(Rsq, polfac[i]*polfac[j],aCC, ts0, ts1);

                phi[i] += ts0*charge[j];
                phi[j] += ts0*charge[i];
                for (size_t k = 0; k < 3; ++k) {
                    Efq[i3 + k] += ts1*charge[j]*Rij[k];
                    Efq[j3 + k] -= ts1*charge[i]*Rij[k];
                }
            }

            // dipole-dipole tensor

            double aDD = ij_bonded ?
             (ij_12 ? smr.aDD_intra_12() : (ij_13 ? smr.aDD_intra_13() 
                                         : smr.aDD_intra_14()))
                                         : smr.aDD_inter();

            if ( !ij_bonded) {
              if (n_i_vector[i] == "n" && n_i_vector[j] == "n") {
                aDD = DDnn;
              } else if (n_i_vector[i] == "n" || n_i_vector[j] == "n"){
                aDD = DDni;
              } else if (n_i_vector[i] == "i" || n_i_vector[j] == "i"){
                aDD = DDii;
              } else {
                throw std::logic_error("molecule type not defined");
              }
            }



            double ts1, ts2;
            smr.smear2(Rsq, polfac[i]*polfac[j], aDD, ts1, ts2);

            double dd3[3][3];

            dd3[0][0] = 3.0*ts2*Rij[0]*Rij[0] - ts1;
            dd3[1][1] = 3.0*ts2*Rij[1]*Rij[1] - ts1;
            dd3[2][2] = 3.0*ts2*Rij[2]*Rij[2] - ts1;
            dd3[0][1] = 3.0*ts2*Rij[0]*Rij[1];
            dd3[0][2] = 3.0*ts2*Rij[0]*Rij[2];
            dd3[1][2] = 3.0*ts2*Rij[1]*Rij[2];
            dd3[1][0] = dd3[0][1];
            dd3[2][0] = dd3[0][2];
            dd3[2][1] = dd3[1][2];

            const double aiaj = polar_sqrt[i3]*polar_sqrt[j3];
            for (size_t k = 0; k < 3; ++k)
                for (size_t l = 0; l < 3; ++l)
                    ddt[3*natom*(i3 + k) + j3 + l] = -aiaj*dd3[k][l];
        }
    }

    // populate the diagonal
    for (size_t i = 0; i < 3*natom; ++i)
        ddt[3*natom*i + i] = 1.0;

    // perform Cholesky decomposition ddt = L*L^T
    // storing L in the lower triangle of ddt and
    // its diagonal in Efd

    const size_t natom3 = 3*natom;
    for (size_t i = 0; i < natom3; ++i) {
        for (size_t j = i; j < natom3; ++j) {

            double sum = ddt[i*natom3 + j];
            for (size_t k = 0; k < i; ++k)
                sum -= ddt[i*natom3 + k]*ddt[j*natom3 + k];

            if(i == j) {
                if (!(sum > 0.0))
                    throw std::logic_error("induced dipoles diverged");
                Efd[i] = std::sqrt(sum);
            } else {
                ddt[j*natom3 + i] = sum/Efd[i];
            }
        }
    }

    // solve L*y = sqrt(a)*Efq storing y in dipole[]
    for (size_t i = 0; i < natom3; ++i) {
        double sum = polar_sqrt[i]*Efq[i];
        for (size_t k = 0; k < i; ++k)
            sum -= ddt[i*natom3 + k]*dipole[k];

        dipole[i] = sum/Efd[i];
    }

    // solve L^T*x = y
    for (size_t i = natom3; i > 0; --i) {
        const size_t i1 = i - 1;
        double sum = dipole[i1];
        for (size_t k = i; k < natom3; ++k)
            sum -= ddt[k*natom3 + i1]*dipole[k];

        dipole[i1] = sum/Efd[i1];
    }

    for (size_t i = 0; i < natom3; ++i)
        dipole[i] *= polar_sqrt[i];

    double Eelec = 0.0;
    for (size_t i = 0; i < natom; ++i)
        Eelec += phi[i]*charge[i];

    Eelec *= 0.5;

    double Eind = 0.0;
    for (size_t i = 0; i < 3*natom; ++i)
        Eind -= dipole[i]*Efq[i];

    Eind *= 0.5;

    m_Eind = Eind;
    m_Eelec = Eelec;

#   ifdef DEBUG
    double U_spring(0);
    for (size_t n = 0; n < natom; ++n) {
        if (polar[n] > 0.0)
            for (size_t k = 0; k < 3; ++k)
                U_spring += dipole[3*n + k]*dipole[3*n + k]/polar[n];
    }

    U_spring /= 2;

    std::cerr << "Eelec = " << Eelec << std:: endl
              << "Eind  = " << Eind << std::endl;
#   endif // DEBUG

    // that is all if no gradients requested
    if (grad == 0)
        return Eelec + Eind;

    // charge-charge interactions
    for (size_t i = 0; i < natom; ++i)
        for (size_t k = 0; k < 3; ++k)
            grad[3*i + k] = -charge[i]*Efq[3*i + k];

#   ifdef DEBUG
    double U_CD(0);
    double U_DD(0);
#   endif // DEBUG

    // charge-dipole interactions
    for (size_t i = 0; i < natom; ++i) {
        const size_t i3 = 3*i;
        for (size_t j = 0; j < natom; ++j) {
            const size_t j3 = 3*j;
            const double qj = charge[j];

            const bool skip_ij = (i == j)
                               || (i < j ? ( is_excluded(i, j, excluded12)
                                         ||  is_excluded(i, j, excluded13)
                                         ||  is_excluded(i, j, excluded14))
                                         : ( is_excluded(j, i, excluded12)
                                         ||  is_excluded(j, i, excluded13)
                                         ||  is_excluded(j, i, excluded14)));

            if (skip_ij)
                continue; // skip this (i, j) pair

            double Rij[3], Rsq(0), diR(0);
            for (size_t k = 0; k < 3; ++k) {
                Rij[k] = xyz[i3 + k] - xyz[j3 + k];
                Rsq += Rij[k]*Rij[k];
                diR += dipole[i3 + k]*Rij[k];
            }

            double ts1, ts2;
            double aCD = smr.aCD();

            if (n_i_vector[i] == "n" && n_i_vector[j] == "n") {
              aCD = CDnn;
            } else if (n_i_vector[i] == "n" || n_i_vector[j] == "n"){
              aCD = CDni;
            } else if (n_i_vector[i] == "i" || n_i_vector[j] == "i"){
              aCD = CDii;
            } else {
              throw std::logic_error("molecule type not defined");
            }

            smr.smear2(Rsq, polfac[i]*polfac[j], smr.aCD(), ts1, ts2);

            for (size_t k = 0; k < 3; ++k) {
                const double derij =
                    qj*(3*ts2*diR*Rij[k] - ts1*dipole[i3 + k]);

                grad[i3 + k] += derij;
                grad[j3 + k] -= derij;
            }

            phi[j] -= ts1*diR;

#           ifdef DEBUG
            U_CD -= qj*diR*ts1; // sign is due to Rij
#           endif // DEBUG

        }
    }

    // dipole-dipole interactions
    for (size_t i = 0; i < natom; ++i) {
        const size_t i3 = 3*i;
        for (size_t j = i + 1; j < natom; ++j) {
            const size_t j3 = 3*j;

            double Rij[3], Rsq(0), diR(0), djR(0), didj(0);
            for (size_t k = 0; k < 3; ++k) {
                Rij[k] = xyz[i3 + k] - xyz[j3 + k];
                Rsq += Rij[k]*Rij[k];
                diR += dipole[i3 + k]*Rij[k];
                djR += dipole[j3 + k]*Rij[k];
                didj += dipole[i3 + k]*dipole[j3 + k];
            }

            const bool ij_12 = is_excluded(i, j, excluded12);
            const bool ij_13 = is_excluded(i, j, excluded13);
            const bool ij_14 = is_excluded(i, j, excluded14);

            const bool ij_bonded = ij_12 || ij_13 || ij_14;

            double aDD = ij_bonded ?
                               (ij_12 ? smr.aDD_intra_12() : 
                               (ij_13 ? smr.aDD_intra_13() : smr.aDD_intra_14()))
                                                           : smr.aDD_inter();

            if ( !ij_bonded) {
              if (n_i_vector[i] == "n" && n_i_vector[j] == "n") {
                aDD = DDnn;
              } else if (n_i_vector[i] == "n" || n_i_vector[j] == "n"){
                aDD = DDni;
              } else if (n_i_vector[i] == "i" || n_i_vector[j] == "i"){
                aDD = DDii;
              } else {
                throw std::logic_error("molecule type not defined");
              }
            }

            double ts1, ts2, ts3;
            smr.smear3(Rsq, polfac[i]*polfac[j], aDD, ts1, ts2, ts3);

            for (size_t k = 0; k < 3; ++k) {
                const double derij =
                   - 3*ts2*(didj*Rij[k] + djR*dipole[i3 + k]
                                        + diR*dipole[j3 + k])
                   + 15*ts3*diR*djR*Rij[k];

                grad[i3 + k] += derij;
                grad[j3 + k] -= derij;
            }
#           ifdef DEBUG
            U_DD += didj*ts1 - 3*diR*djR*ts2;
#           endif // DEBUG
        }
    }

    // Redistribute forces from virtual sites to atomic sites
    for(size_t n = 0; n < molec.size(); ++n){
        if(molec[n]->has_virtual_sites()){
            size_t shift = first_index_of_molec[n];

            molec[n]->redistrib_frc_virtual_2_real(grad + 3*shift);
        }
    }

    // Add force arising from geometry dependence of charges
    for(size_t n = 0; n < molec.size(); ++n){
        if(molec[n]->has_charge_derivatives()){
            size_t shift = first_index_of_molec[n];

            molec[n]->charge_derivative_force(phi + shift, grad + 3*shift);
        }
    }

#   ifdef DEBUG
    std::cerr << "    U_CC = " << Eelec << std::endl;
    std::cerr << "    U_CD = " << U_CD << std::endl;
    std::cerr << "    U_DD = " << U_DD << std::endl;
    std::cerr << "U_spring = " << U_spring << std::endl;

    std::cerr << "\nU_CD + U_DD + U_spring = "
              << (U_CD + U_DD + U_spring) << std::endl;
    std::cerr << "E_ind = " << Eind << std::endl;
#   endif // DEBUG

    return Eelec + Eind;
}


// Original Electrostatics
double electrostatics::operator()(size_t natom, const double* charge,
    const double* polfac, const double* polar, const double* xyz,
    const excluded_set_type& excluded12,
    const excluded_set_type& excluded13,
    const std::vector<int>& first_index_of_molec,
    const std::vector<std::shared_ptr<molecule>>& molec, const smear& smr,
    double* grad)
{
    allocate(natom);

    double* dipole      = m_memory;

    double* polar_sqrt  = dipole + 3*natom;
    double* phi         = polar_sqrt + 3*natom;
    double* Efq         = phi + natom;
    double* Efd         = Efq + 3*natom;
    double* ddt         = Efd + 3*natom;

    // FIXME: This is the wrong length?
    std::fill(phi, phi + natom*(1 + 2*3 + 3*3*natom), 0.0);

    // Necessary to check if the ADD must be aDD-nn, -ni or -ii.
    size_t nmol = 0;
    size_t count = 0;
    std::vector<std::string> n_i_vector;
    for (size_t i = 0; i < natom; ++i) {
      if (count >= molec[nmol]->get_nsites()) {
        nmol++;
        count = 0;
      }
      if (molec[nmol]->id == "n") {
        n_i_vector.push_back("n");
      } else if (molec[nmol]->id == "i") {
        n_i_vector.push_back("i");
      }  
      count++;
    }

    for (size_t i = 0; i < natom; ++i) {
        const size_t i3 = 3*i;
        const double sqrtalpha = std::sqrt(polar[i]);

        for (size_t k = 0; k < 3; ++k)
            polar_sqrt[i3 + k] = sqrtalpha;
    }

    for (size_t i = 0; i < natom; ++i) {
        const size_t i3 = 3*i;

        for (size_t j = i + 1; j < natom; ++j) {
            const size_t j3 = 3*j;

            double Rij[3], Rsq(0);
            for (size_t k = 0; k < 3; ++k) {
                Rij[k] = xyz[i3 + k] - xyz[j3 + k];
                Rsq += Rij[k]*Rij[k];
            }

            // charge-charge

            const bool ij_12 = is_excluded(i, j, excluded12);
            const bool ij_13 = is_excluded(i, j, excluded13);

            const bool ij_bonded = ij_12 || ij_13;

            if (!ij_bonded) {
                double ts0, ts1;
                double aCC =  smr.aCC();
                if ( !ij_bonded) {
                  if (n_i_vector[i] == "n" && n_i_vector[j] == "n") {
                    aCC = CCnn;
                  } else if (n_i_vector[i] == "n" || n_i_vector[j] == "n"){
                    aCC = CCni;
                  } else if (n_i_vector[i] == "i" || n_i_vector[j] == "i"){
                    aCC = CCii;
                  } else {
                    throw std::logic_error("molecule type not defined");
                  }
                }

                smr.smear01(Rsq, polfac[i]*polfac[j],aCC, ts0, ts1);

                phi[i] += ts0*charge[j];
                phi[j] += ts0*charge[i];
                for (size_t k = 0; k < 3; ++k) {
                    Efq[i3 + k] += ts1*charge[j]*Rij[k];
                    Efq[j3 + k] -= ts1*charge[i]*Rij[k];
                }
            }

            // dipole-dipole tensor

            double aDD = ij_bonded ?
             (ij_12 ? smr.aDD_intra_12() : smr.aDD_intra_13())
                                         : smr.aDD_inter();

            if ( !ij_bonded) {
              if (n_i_vector[i] == "n" && n_i_vector[j] == "n") {
                aDD = DDnn;
              } else if (n_i_vector[i] == "n" || n_i_vector[j] == "n"){
                aDD = DDni;
              } else if (n_i_vector[i] == "i" || n_i_vector[j] == "i"){
                aDD = DDii;
              } else {
                throw std::logic_error("molecule type not defined");
              }
            }
              
            

            double ts1, ts2;
            smr.smear2(Rsq, polfac[i]*polfac[j], aDD, ts1, ts2);

            double dd3[3][3];

            dd3[0][0] = 3.0*ts2*Rij[0]*Rij[0] - ts1;
            dd3[1][1] = 3.0*ts2*Rij[1]*Rij[1] - ts1;
            dd3[2][2] = 3.0*ts2*Rij[2]*Rij[2] - ts1;
            dd3[0][1] = 3.0*ts2*Rij[0]*Rij[1];
            dd3[0][2] = 3.0*ts2*Rij[0]*Rij[2];
            dd3[1][2] = 3.0*ts2*Rij[1]*Rij[2];
            dd3[1][0] = dd3[0][1];
            dd3[2][0] = dd3[0][2];
            dd3[2][1] = dd3[1][2];

            const double aiaj = polar_sqrt[i3]*polar_sqrt[j3];
            for (size_t k = 0; k < 3; ++k)
                for (size_t l = 0; l < 3; ++l)
                    ddt[3*natom*(i3 + k) + j3 + l] = -aiaj*dd3[k][l];
        }
    }


    // populate the diagonal
    for (size_t i = 0; i < 3*natom; ++i)
        ddt[3*natom*i + i] = 1.0;

    // perform Cholesky decomposition ddt = L*L^T
    // storing L in the lower triangle of ddt and
    // its diagonal in Efd

    const size_t natom3 = 3*natom;
    for (size_t i = 0; i < natom3; ++i) {
        for (size_t j = i; j < natom3; ++j) {

            double sum = ddt[i*natom3 + j];
            for (size_t k = 0; k < i; ++k)
                sum -= ddt[i*natom3 + k]*ddt[j*natom3 + k];

            if(i == j) {
                if (!(sum > 0.0))
                    throw std::logic_error("induced dipoles diverged");
                Efd[i] = std::sqrt(sum);
            } else {
                ddt[j*natom3 + i] = sum/Efd[i];
            }
        }
    }

    // solve L*y = sqrt(a)*Efq storing y in dipole[]
    for (size_t i = 0; i < natom3; ++i) {
        double sum = polar_sqrt[i]*Efq[i];
        for (size_t k = 0; k < i; ++k)
            sum -= ddt[i*natom3 + k]*dipole[k];

        dipole[i] = sum/Efd[i];
    }

    // solve L^T*x = y
    for (size_t i = natom3; i > 0; --i) {
        const size_t i1 = i - 1;
        double sum = dipole[i1];
        for (size_t k = i; k < natom3; ++k)
            sum -= ddt[k*natom3 + i1]*dipole[k];

        dipole[i1] = sum/Efd[i1];
    }

    for (size_t i = 0; i < natom3; ++i)
        dipole[i] *= polar_sqrt[i];

    double Eelec = 0.0;
    for (size_t i = 0; i < natom; ++i)
        Eelec += phi[i]*charge[i];

    Eelec *= 0.5;

    double Eind = 0.0;
    for (size_t i = 0; i < 3*natom; ++i)
        Eind -= dipole[i]*Efq[i];

    Eind *= 0.5;

    m_Eind = Eind;
    m_Eelec = Eelec;

#   ifdef DEBUG
    double U_spring(0);
    for (size_t n = 0; n < natom; ++n) {
        if (polar[n] > 0.0)
            for (size_t k = 0; k < 3; ++k)
                U_spring += dipole[3*n + k]*dipole[3*n + k]/polar[n];
    }

    U_spring /= 2;

    std::cerr << "Eelec = " << Eelec << std:: endl
              << "Eind  = " << Eind << std::endl;
#   endif // DEBUG

    // that is all if no gradients requested
    if (grad == 0)
        return Eelec + Eind;

    // charge-charge interactions
    for (size_t i = 0; i < natom; ++i)
        for (size_t k = 0; k < 3; ++k)
            grad[3*i + k] = -charge[i]*Efq[3*i + k];

#   ifdef DEBUG
    double U_CD(0);
    double U_DD(0);
#   endif // DEBUG

    // charge-dipole interactions
    for (size_t i = 0; i < natom; ++i) {
        const size_t i3 = 3*i;
        for (size_t j = 0; j < natom; ++j) {
            const size_t j3 = 3*j;
            const double qj = charge[j];

            const bool skip_ij = (i == j)
                               || (i < j ? ( is_excluded(i, j, excluded12)
                                         ||  is_excluded(i, j, excluded13))
                                         : ( is_excluded(j, i, excluded12)
                                         ||  is_excluded(j, i, excluded13)));

            if (skip_ij)
                continue; // skip this (i, j) pair

            double Rij[3], Rsq(0), diR(0);
            for (size_t k = 0; k < 3; ++k) {
                Rij[k] = xyz[i3 + k] - xyz[j3 + k];
                Rsq += Rij[k]*Rij[k];
                diR += dipole[i3 + k]*Rij[k];
            }

            double ts1, ts2;
            double aCD = smr.aCD();

            if (n_i_vector[i] == "n" && n_i_vector[j] == "n") {
              aCD = CDnn;
            } else if (n_i_vector[i] == "n" || n_i_vector[j] == "n"){
              aCD = CDni;
            } else if (n_i_vector[i] == "i" || n_i_vector[j] == "i"){
              aCD = CDii;
            } else {
              throw std::logic_error("molecule type not defined");
            }

            smr.smear2(Rsq, polfac[i]*polfac[j], smr.aCD(), ts1, ts2);

            for (size_t k = 0; k < 3; ++k) {
                const double derij =
                    qj*(3*ts2*diR*Rij[k] - ts1*dipole[i3 + k]);

                grad[i3 + k] += derij;
                grad[j3 + k] -= derij;
            }

            phi[j] -= ts1*diR;

#           ifdef DEBUG
            U_CD -= qj*diR*ts1; // sign is due to Rij
#           endif // DEBUG

        }
    }

    // dipole-dipole interactions
    for (size_t i = 0; i < natom; ++i) {
        const size_t i3 = 3*i;
        for (size_t j = i + 1; j < natom; ++j) {
            const size_t j3 = 3*j;

            double Rij[3], Rsq(0), diR(0), djR(0), didj(0);
            for (size_t k = 0; k < 3; ++k) {
                Rij[k] = xyz[i3 + k] - xyz[j3 + k];
                Rsq += Rij[k]*Rij[k];
                diR += dipole[i3 + k]*Rij[k];
                djR += dipole[j3 + k]*Rij[k];
                didj += dipole[i3 + k]*dipole[j3 + k];
            }

            const bool ij_12 = is_excluded(i, j, excluded12);
            const bool ij_13 = is_excluded(i, j, excluded13);

            const bool ij_bonded = ij_12 || ij_13;

            double aDD = ij_bonded ?
                               (ij_12 ? smr.aDD_intra_12() : smr.aDD_intra_13())
                                                           : smr.aDD_inter();

            if ( !ij_bonded) {
              if (n_i_vector[i] == "n" && n_i_vector[j] == "n") {
                aDD = DDnn;
              } else if (n_i_vector[i] == "n" || n_i_vector[j] == "n"){
                aDD = DDni;
              } else if (n_i_vector[i] == "i" || n_i_vector[j] == "i"){
                aDD = DDii;
              } else {
                throw std::logic_error("molecule type not defined");
              }
            }

            double ts1, ts2, ts3;
            smr.smear3(Rsq, polfac[i]*polfac[j], aDD, ts1, ts2, ts3);

            for (size_t k = 0; k < 3; ++k) {
                const double derij =
                   - 3*ts2*(didj*Rij[k] + djR*dipole[i3 + k]
                                        + diR*dipole[j3 + k])
                   + 15*ts3*diR*djR*Rij[k];

                grad[i3 + k] += derij;
                grad[j3 + k] -= derij;
            }
#           ifdef DEBUG
            U_DD += didj*ts1 - 3*diR*djR*ts2;
#           endif // DEBUG
        }
    }

    // Redistribute forces from virtual sites to atomic sites
    for(size_t n = 0; n < molec.size(); ++n){
        if(molec[n]->has_virtual_sites()){
            size_t shift = first_index_of_molec[n];

            molec[n]->redistrib_frc_virtual_2_real(grad + 3*shift);
        }
    }

    // Add force arising from geometry dependence of charges
    for(size_t n = 0; n < molec.size(); ++n){
        if(molec[n]->has_charge_derivatives()){
            size_t shift = first_index_of_molec[n];

            molec[n]->charge_derivative_force(phi + shift, grad + 3*shift);
        }
    }

#   ifdef DEBUG
    std::cerr << "    U_CC = " << Eelec << std::endl;
    std::cerr << "    U_CD = " << U_CD << std::endl;
    std::cerr << "    U_DD = " << U_DD << std::endl;
    std::cerr << "U_spring = " << U_spring << std::endl;

    std::cerr << "\nU_CD + U_DD + U_spring = "
              << (U_CD + U_DD + U_spring) << std::endl;
    std::cerr << "E_ind = " << Eind << std::endl;
#   endif // DEBUG

    return Eelec + Eind;
}

//----------------------------------------------------------------------------//

//
// assumes that () was called
//

void electrostatics::polarizability_tensor(size_t natom, double* pol)
{
    // works only with Cholesky decomposition

    double* dipole      = m_memory;
    double* polar_sqrt  = dipole + 3*natom;
    double* phi         = polar_sqrt + 3*natom;
    double* Efq         = phi + natom;
    double* Efd         = Efq + 3*natom;
    double* ddt         = Efd + 3*natom;

    // ddt = L*L^T [done in operator()]
    // L in the lower triangle of ddt and
    // its diagonal in Efd

    // we need B = alpha^{1/2}*ddt^{-1}*alpha^{1/2}

    // compute L^{-1} following NR (ch 2.9)

    const size_t natom3 = 3*natom;

    for (size_t i = 0; i < natom3; ++i) {
        ddt[i*natom3 + i] = 1.0/Efd[i];

        for (size_t j = i + 1; j < natom3; ++j) {

            double sum(0);
            for (size_t k = i; k < j; ++k)
                sum -= ddt[j*natom3 + k]*ddt[k*natom3 + i];

            ddt[j*natom3 + i] = sum/Efd[j];
        }
    }

    std::fill(pol, pol + 9, 0.0);

    for (size_t i = 0; i < natom; ++i) {
        for (size_t j = 0; j < natom; ++j) {
            for (size_t a = 0; a < 3; ++a) {
                for (size_t b = 0; b < 3; ++b) {

                    double sum(0);

                    const size_t ia = 3*i + a;
                    const size_t jb = 3*j + b;

                    for (size_t k = std::max(ia, jb); k < natom3; ++k)
                        sum += ddt[k*natom3 + ia]*ddt[k*natom3 + jb];

                    pol[3*a + b] += sum*polar_sqrt[ia]*polar_sqrt[jb];
                }
            }
        }
    }
}

//----------------------------------------------------------------------------//

} // namespace ttm

////////////////////////////////////////////////////////////////////////////////
