#include <iostream>
#include <utility>
#include <cstddef>

#include "molecule.h"

//------------------  Constructors / Destructors / Memory Allocation  --------//

// constructor
molecule::molecule()
{
    definedNsites = false;
    definedCoords = false;
    definedCharges = false;
    definedPolarizability = false;
};

// destructor
molecule::~molecule()
{
    delete[] memory;
}

// copy constructor
molecule::molecule(const molecule& o){

    // Allocate memory and copy values
    allocate(o.nsites);
    std::copy(o.memory, o.memory + 6*o.nsites, memory);

    // Set pointers
    charge   = memory;
    coords   = memory + nsites;
    atmpolar = memory + nsites + nsites*3;
    polfac   = memory + nsites + nsites*3 + nsites;
    definedCoords = definedCharges = definedPolarizability = true;

    id = o.id;
}

// sets nsites and allocates memory
void molecule::allocate(const int n){
    nsites = n;
    definedNsites = true;

    memory = new double [ nsites    //charge
        + nsites*3  //site coordinates
        + nsites    //polarizabilities
        + nsites];  //polfac
}

//------------------  Functions for setting data in base class ---------------//

// sets site positions based on atom coordinates
void molecule::set_coords(double* xyz){
    coords = memory + nsites;

    for(size_t i = 0; i < 3*nsites; ++i)
        coords[i] = xyz[i];

    definedCoords = true;
    virtual_sites = false;
}

// default (virtual) set_charges assumes charges are geometry independent
// this set_charges is overridden in derived classes for molecules which
// have a geometry dependences for their charges
void molecule::set_charges(double* q){
    charge = memory;

    for(size_t i = 0; i < nsites; ++i)
        charge[i] = q[i]*constants::CHARGECON;

    definedCharges = true;
    charge_derivatives = false;
}

// sets polarizabilities
void molecule::set_pol(double* p){
    atmpolar = memory + nsites + nsites*3;
    polfac = memory + nsites + nsites*3 + nsites;

    for(size_t i = 0; i < nsites; ++i){
        atmpolar[i] = p[i];
    }

    std::copy(atmpolar, atmpolar + nsites, polfac);

    definedPolarizability = true;
}

void molecule::set_polfac(double* p){
    polfac = memory + nsites + nsites*3 + nsites;

    for(size_t i = 0; i < nsites; ++i){
        polfac[i] = p[i];
    }

}

//------------------  Functions for accessing data in base class -------------//

size_t molecule::get_nsites()
{
    if(definedNsites)
        return nsites;
    else{
        std::cerr << "nsites have not been defined" << std::endl;
        exit(1);
    }
}

double* molecule::get_charges()
{
    if(definedCharges)
        return charge;
    else{
        std::cerr << "charges have not been defined" << std::endl;
        exit(1);
    }
}

double* molecule::get_coords()
{
    if(definedCoords)
        return coords;
    else{
        std::cerr << "coords have not been defined" << std::endl;
        exit(1);
    }
}

double* molecule::get_pol()
{
    if(definedPolarizability)
        return atmpolar;
    else{
        std::cerr << "Polarizabilities have not been defined" << std::endl;
        exit(1);
    }
}

double* molecule::get_polfac()
{
    if(definedPolarizability)
        return polfac;
    else{
        std::cerr << "polfacs have not been defined" << std::endl;
        exit(1);
    }
}

void molecule::get_charges(double* new_charges)
{
    if(definedCharges)
        std::copy(charge, charge + nsites, new_charges);
    else{
        std::cerr << "charges have not been defined" << std::endl;
        exit(1);
    }
}

void molecule::get_coords(double* new_coords)
{
    if(definedCoords)
        std::copy(coords, coords + 3*nsites, new_coords);
    else{
        std::cerr << "coords have not been defined" << std::endl;
        exit(1);
    }
}

void molecule::get_pol(double* new_atmpolar)
{
    if(definedPolarizability)
        std::copy(atmpolar, atmpolar + nsites, new_atmpolar);
    else{
        std::cerr << "Polarizabilities have not been defined" << std::endl;
        exit(1);
    }
}

void molecule::get_polfac(double* new_polfac)
{
    if(definedPolarizability)
        std::copy(polfac, polfac + nsites, new_polfac);
    else{
        std::cerr << "polfacs have not been defined" << std::endl;
        exit(1);
    }
}

bool molecule::has_virtual_sites()
{
    return virtual_sites;
}

bool molecule::has_charge_derivatives()
{
    return charge_derivatives;
}

//----------------------------------------------------------------------------//
