#ifdef HAVE_CONFIG_H
#   include "config.h"
#endif // HAVE_CONFIG_H

#include <cmath>
#include <cassert>

#include <iostream>
#include <cstdlib>

#include "pot-water-ion.h"

#include "ps.h"
#include "x2b-v9x.h"
#include "x3b-v2x.h"
#include "x2b-dispersion.h"

#include "watI-dispersion.h"
#include "watI-buckingham.h"
#include "II_Dispersion.h"
#include "II_Buckingham.h"
#include "II_Poly.h"
#include "orient.h"

//#define VERBOSE 
//#define PRINT_INTERACTION_ENERGY
////////////////////////////////////////////////////////////////////////////////

namespace {

const double E1max = 100.0; // kcal/mol

double m_charge = 0.0;
double m_polar  = 0.0;
double m_cOX    = 0.0;
double m_dOX    = 0.0;
double m_aOX    = 0.0;
double m_bOX    = 0.0;
double m_cHX    = 0.0;
double m_dHX    = 0.0;
double m_aHX    = 0.0;
double m_bHX    = 0.0;

bool params_defined = false;
bool params_set     = false;

}

////////////////////////////////////////////////////////////////////////////////

namespace x2o {

void pot_water_ion::define_ion_params(const double* params)
{

    m_charge = params[0];
    m_polar  = params[1];
    m_cOX    = params[2];
    m_dOX    = params[3];
    m_aOX    = params[4];
    m_bOX    = params[5];
    m_cHX    = params[6];
    m_dHX    = params[7];
    m_aHX    = params[8];
    m_bHX    = params[9];

    params_defined = true;

}

//----------------------------------------------------------------------------//

double pot_water_ion::get_E_water(size_t nw, const double* pos)
{
  assert(nw > 0 && pos);
  double E1(0);

  bool good = true;

  // 1-body : just water
  for (size_t i = 0; i < nw; ++i) {
    const size_t i9 = 9*i;

    const double this_E1 = ps::pot_nasa(pos + i9, 0);
    E1 += this_E1;
    if (this_E1 > E1max)
      good = false;
  }

  double  Edisp(0), E2poly(0), E3poly(0);

  if (good) {
    for (size_t i = 0; i < nw; ++i) {
      const size_t i9 = 9*i;
      // Two-body for water-water (dispersion + short-range)
      for (size_t j = i + 1; j < nw; ++j) {
        const size_t j9 = 9*j;
        Edisp += x2b_dispersion::eval(pos + i9, pos + j9);
        E2poly += x2b_v9x::eval(pos + i9, pos + j9);
        // Three-body for water-water-water (short-range)
        for (size_t k = j + 1; k < nw; ++k) {
          const size_t k9 = 9*k;
          E3poly += x3b_v2x::eval(pos + i9, pos + j9, pos + k9);
        }
      }
    } 
  } // good

#ifdef VERBOSE
    std::cout << "\n WATER ALONE \n    E1 = " << E1
              << "\n Edisp = " << Edisp
              << "\nE2poly = " << E2poly
              << "\nE3poly = " << E3poly
              << std::endl; 
#endif

  return E1 +  Edisp + E2poly + E3poly ;
}

double pot_water_ion::get_E_water_1b(size_t nw, const double* pos)
{
  assert(nw > 0 && pos);
  double E1(0);

  // 1-body : just water
  for (size_t i = 0; i < nw; ++i) {
    const size_t i9 = 9*i;

    const double this_E1 = ps::pot_nasa(pos + i9, 0);
    E1 += this_E1;
  }

#ifdef VERBOSE
    std::cout << "\n WATER ALONE \n    E1 = " << E1
              << std::endl;
#endif

  return E1 ;
}

double pot_water_ion::get_E_water(size_t nw, const double* pos, double* grd)
{
  assert(nw > 0 && pos);
  double E1(0);

  bool good = true;
  std::fill(grd, grd + 3*3*nw, 0.0);

  // 1-body : just water
  for (size_t i = 0; i < nw; ++i) {
    const size_t i9 = 9*i;

    const double this_E1 = ps::pot_nasa(pos + i9, grd + i9);
    E1 += this_E1;
    if (this_E1 > E1max)
      good = false;
  }

  double  Edisp(0), E2poly(0), E3poly(0);

  if (good) {
    for (size_t i = 0; i < nw; ++i) {
      const size_t i9 = 9*i;
      // Two-body for water-water (dispersion + short-range)
      for (size_t j = i + 1; j < nw; ++j) {
        const size_t j9 = 9*j;
        E2poly += x2b_v9x::eval(pos + i9, pos + j9, grd + i9, grd + j9);
        Edisp += x2b_dispersion::eval(pos + i9, pos + j9,grd + i9, grd + j9);
        // Three-body for water-water-water (short-range)
        for (size_t k = j + 1; k < nw; ++k) {
          const size_t k9 = 9*k;
          E3poly += x3b_v2x::eval(pos + i9, pos + j9, pos + k9, grd + i9, grd + j9, grd + k9);
        }
      }
    }
  } // good

  

#ifdef VERBOSE
    std::cout << "\n WATER ALONE \n    E1 = " << E1
              << "\n Edisp = " << Edisp
              << "\nE2poly = " << E2poly
              << "\nE3poly = " << E3poly
              << std::endl;
#endif

  return E1 +  Edisp + E2poly + E3poly ;
}


double pot_water_ion::get_E_water_ion(size_t nw, size_t nI, const double* pos){
  assert(nw > 0 && pos);
  double E1(0);
  
  watI_dispersion m_watI_dispersion;
  watI_buckingham m_watI_buckingham;
  // Setting parameters for water-ion
  if(nI != 0){
      m_watI_buckingham.set_b_params(m_aOX, m_bOX, m_aHX, m_bHX);
      m_watI_dispersion.set_tt_params(m_cOX, m_dOX, m_cHX, m_dHX);

      params_set = true;
  }else{
      std::cerr << "No ions defined" << std::endl;
      exit(1);
    
  }

  bool good = true;

  // 1-body : just water
  for (size_t i = 0; i < nw; ++i) {
    const size_t i9 = 9*i;

    const double this_E1 = ps::pot_nasa(pos + i9, 0);
    E1 += this_E1;
    if (this_E1 > E1max)
      good = false;
  }

  double Edisp(0), E2wat_ion(0);

  if (good) {
    for (size_t i = 0; i < nw; ++i) {
      const size_t i9 = 9*i;
      for (size_t j = 0; j < nI; ++j) {
        const size_t nw9j = nw*9 + 3*j;
        Edisp += m_watI_dispersion(pos + i9, pos + nw9j);
        E2wat_ion += m_watI_buckingham(pos + i9, pos + nw9j);
      }
    }
  } // good

#ifdef VERBOSE
    std::cout << "\n WATER - ION \n" << "\n Edisp = " << Edisp
              << "\nE_buck = " << E2wat_ion
              << std::endl;
#endif

  return Edisp + E2wat_ion;
}

double pot_water_ion::get_E_water_ion(size_t nw, size_t nI, const double* pos, double * grd){
  assert(nw > 0 && pos);
  double E1(0);

  watI_dispersion m_watI_dispersion;
  watI_buckingham m_watI_buckingham;
  // Setting parameters for water-ion
  if(nI != 0){
      m_watI_buckingham.set_b_params(m_aOX, m_bOX, m_aHX, m_bHX);
      m_watI_dispersion.set_tt_params(m_cOX, m_dOX, m_cHX, m_dHX);

  }else{
      std::cerr << "Ion not defined" << std::endl;
      exit(1);
  }

  std::fill(grd, grd + 3*3*nw + 3*nI, 0.0);
  bool good = true;

  // 1-body : just water
  for (size_t i = 0; i < nw; ++i) {
    const size_t i9 = 9*i;

    const double this_E1 = ps::pot_nasa(pos + i9, 0);
    E1 += this_E1;
    if (this_E1 > E1max)
      good = false;
  }

  double Edisp(0), E2wat_ion(0);

  if (good) {
    for (size_t i = 0; i < nw; ++i) {
      const size_t i9 = 9*i;
      for (size_t j = 0; j < nI; ++j) {
        const size_t nw9j = nw*9 + 3*j;
        Edisp += m_watI_dispersion(pos + i9, pos + nw9j, grd + i9, grd + nw9j);
        E2wat_ion += m_watI_buckingham(pos + i9, pos + nw9j, grd + i9, grd + nw9j);
      }
    }
  } // good

#ifdef VERBOSE
    std::cout << "\n WATER - ION \n" << "\n Edisp = " << Edisp
              << "\nE_buck = " << E2wat_ion
              << std::endl;
#endif


  return Edisp + E2wat_ion;
}

double pot_water_ion::get_E_ion_ion(size_t nI1, size_t nI2, const double* pos) {

  assert(nI1 > 0 && pos);

  II_Dispersion m_II_dispersion;
  II_Buckingham m_II_buckingham;

  if(nI2 != 0){
      m_II_dispersion.set_params(m_cOX, m_dOX);
      m_II_buckingham.set_params(m_aOX, m_bOX);

  }else{
      std::cerr << "Ion not defined" << std::endl;
      exit(1);
  }

  bool good = true;

  double Edisp(0), E2wat_ion(0);

  if (good) {
    for (size_t i = 0; i < nI1; ++i) {
      const size_t i3 = 3*i;
      for (size_t j = 0; j < nI2; ++j) {
        const size_t nI13j = nI1*3 + 3*j;
        Edisp += m_II_dispersion(pos + i3, pos + nI13j);
        E2wat_ion += m_II_buckingham(pos + i3, pos + nI13j);
      }
    }
  } // good

#ifdef VERBOSE
    std::cout << "\n ION - ION \n" << "\n Edisp = " << Edisp
              << "\nE_buck = " << E2wat_ion
              << std::endl;
#endif



  return Edisp + E2wat_ion;


}

double pot_water_ion::get_E_ion_ion(size_t nI1, size_t nI2, const double* pos, double * grd) {

  assert(nI1 > 0 && pos);

  std::fill(grd, grd + 3*nI1 + 3*nI2, 0.0);
  II_Dispersion m_II_dispersion;
  II_Buckingham m_II_buckingham;

  if(nI2 != 0){
      m_II_dispersion.set_params(m_cOX, m_dOX);
      m_II_buckingham.set_params(m_aOX, m_bOX);

  }else{
      std::cerr << "Ion not defined" << std::endl;
      exit(1);
  }

  bool good = true;

  double Edisp(0), E2wat_ion(0);

  if (good) {
    for (size_t i = 0; i < nI1; ++i) {
      const size_t i3 = 3*i;
      for (size_t j = 0; j < nI2; ++j) {
        const size_t nI13j = nI1*3 + 3*j;
        Edisp += m_II_dispersion(pos + i3, pos + nI13j, grd + i3, grd  + nI13j);
        E2wat_ion += m_II_buckingham(pos + i3, pos + nI13j, grd + i3, grd  + nI13j);
      }
    }
  } // good

#ifdef VERBOSE
    std::cout << "\n ION - ION \n" << "\n Edisp = " << Edisp
              << "\nE_buck = " << E2wat_ion
              << std::endl;
#endif

  return Edisp + E2wat_ion;
}

double pot_water_ion::get_E_ion_ion_mbnrg(size_t nI1, size_t nI2, const double* pos, double * grd) {

  assert(nI1 > 0 && pos);

  std::fill(grd, grd + 3*nI1 + 3*nI2, 0.0);
  II_Dispersion m_II_dispersion;
  II_Poly m_II_Poly;

  if(nI2 != 0){
      m_II_dispersion.set_params(m_cOX, m_dOX);

  }else{
      std::cerr << "Ion not defined" << std::endl;
      exit(1);
  }

  bool good = true;

  double Edisp(0), E2wat_ion(0);

  if (good) {
    for (size_t i = 0; i < nI1; ++i) {
      const size_t i3 = 3*i;
      for (size_t j = 0; j < nI2; ++j) {
        const size_t nI13j = nI1*3 + 3*j;
        Edisp += m_II_dispersion(pos + i3, pos + nI13j, grd + i3, grd  + nI13j);
//        E2wat_ion += m_II_buckingham(pos + i3, pos + nI13j, grd + i3, grd  + nI13j);
      }
    }
  } // good

#ifdef VERBOSE
    std::cout << "\n ION - ION \n" << "\n Edisp = " << Edisp
              << "\nE_buck = " << E2wat_ion
              << std::endl;
#endif

  return Edisp + E2wat_ion;
}

double pot_water_ion::operator()(size_t nw, size_t nI, const double* pos) const
{
    assert(nw > 0 && pos);

    double E1(0);
	// Setting parameters for water-ion
    if(params_set == false && nI != 0){
        if(params_defined == true){
            m_watI_buckingham.set_b_params(m_aOX, m_bOX, m_aHX, m_bHX);
            m_watI_dispersion.set_tt_params(m_cOX, m_dOX, m_cHX, m_dHX);
            m_ttm4.set_ion_params(m_charge, m_polar);

            params_set = true;
       }else{
            std::cerr << "Ion parameters not defined" << std::endl;
            exit(1);
       }
    }


    bool good = true;

    // 1-body : just water
    for (size_t i = 0; i < nw; ++i) {
        const size_t i9 = 9*i;

        const double this_E1 = ps::pot_nasa(pos + i9, 0);
        E1 += this_E1;

        if (this_E1 > E1max)
            good = false;
    }

    double Eelec(0), Eind(0), Edisp(0), E2poly(0), E3poly(0), E2wat_ion(0);

    if (good) {

        m_ttm4(nw, nI, pos, Eelec, 0, Eind, 0);

        std::cerr << "Elec: " << Eelec + Eind << std::endl;

        for (size_t i = 0; i < nw; ++i) {
            const size_t i9 = 9*i;

            // Two-body for water-water (dispersion + short-range)
            for (size_t j = i + 1; j < nw; ++j) {
                const size_t j9 = 9*j;
                Edisp += x2b_dispersion::eval(pos + i9, pos + j9);
                E2poly += x2b_v9x::eval(pos + i9, pos + j9);

                // Three-body for water-water-water (short-range)
                for (size_t k = j + 1; k < nw; ++k) {
                    const size_t k9 = 9*k;
                    E3poly += x3b_v2x::eval(pos + i9, pos + j9, pos + k9);
                }
            }

            // Two-body for water-ion (dispersion + short-range)
            for (size_t j = 0; j < nI; ++j) {
                const size_t nw9j = nw*9 + 3*j;
                // call dispersion (added to the previous one)
                Edisp += m_watI_dispersion(pos + i9, pos + nw9j);
                // call short-range (buckingham)
		E2wat_ion += m_watI_buckingham(pos + i9, pos + nw9j);
            }

        }
    } // good

#   ifdef VERBOSE
    std::cout << "\n    E1 = " << E1
              << "\n Eelec = " << Eelec
              << "\n  Eind = " << Eind
              << "\n Edisp = " << Edisp
              << "\nE2poly = " << E2poly
              << "\nE3poly = " << E3poly
              << "\nE2wat_ion = " << E2wat_ion
              << std::endl;
#   endif /* VERBOSE */

#ifdef PRINT_INTERACTION_ENERGY
    double E_int = Eelec + Eind + Edisp + E2poly + E3poly + E2wat_ion;
    std::cout << std::endl << "Eint = " << E_int;
#endif

    return E1 + Eelec + Eind + Edisp + E2poly + E3poly + E2wat_ion;
}

//----------------------------------------------------------------------------//

double pot_water_ion::operator()(size_t nw, size_t nI, const double* pos, 
				 double* grd) const
{
    assert(nw > 0 && pos && grd);

	// Setting parameters for water-ion
    if(params_set == false && nI != 0){
        if(params_defined == true){
            m_watI_buckingham.set_b_params(m_aOX, m_bOX, m_aHX, m_bHX);
            m_watI_dispersion.set_tt_params(m_cOX, m_dOX, m_cHX, m_dHX);
            m_ttm4.set_ion_params(m_charge, m_polar);

            params_set = true;
       }else{
            std::cerr << "Ion parameters not defined" << std::endl;
            exit(1);
       }
    }

    std::fill(grd, grd + 3*3*nw + 3*nI, 0.0);
    double Etot(0);
    bool good(true);

    
    double tot_E1(0);

    for (size_t i = 0; i < nw; ++i) {
        const size_t i9 = 9*i;

        const double E1 = ps::pot_nasa(pos + i9, grd + i9);
        if (E1 > E1max)
            good = false;

        tot_E1 += E1;
        Etot += E1;
    }

    if (good) {
        double Eelec(0), Eind(0), gEelec[9*nw + 3*nI], gEind[9*nw + 3*nI];

        m_ttm4(nw, nI, pos, Eelec, gEelec, Eind, gEind);
        for (size_t i = 0; i < 9*nw + 3*nI; ++i)
            grd[i] += gEelec[i] + gEind[i];

        Etot += Eelec + Eind;
        std::cerr << "Elec: " << Eelec + Eind << std::endl;

        for (size_t i = 0; i < nw; ++i) {
            const size_t i9 = 9*i;

            for (size_t j = i + 1; j < nw; ++j) {
                const size_t j9 = 9*j;
                Etot += x2b_v9x::eval(pos + i9, pos + j9, grd + i9, grd + j9)
                      + x2b_dispersion::eval(pos + i9, pos + j9,
                                             grd + i9, grd + j9);

                for (size_t k = j + 1; k < nw; ++k) {
                    const size_t k9 = 9*k;
                    Etot += x3b_v2x::eval(pos + i9, pos + j9, pos + k9,
                                          grd + i9, grd + j9, grd + k9);
                }
            }

            // Two-body for water-ion (dispersion + short-range)
            for (size_t j = 0; j < nI; ++j) {
                const size_t nw9j = nw*9 + 3*j;
                // call dispersion (added to the previous one)
                Etot += m_watI_dispersion(pos + i9, pos + nw9j,
	       				  grd + i9, grd + nw9j);
                // call short-range (buckingham)
                Etot += m_watI_buckingham(pos + i9, pos + nw9j,
	                                  grd + i9, grd + nw9j);
            }
        }
    } // good

#ifdef PRINT_INTERACTION_ENERGY
    double E_int = Etot - tot_E1;
    std::cout << "Interaction energy [grd] = " << E_int << std::endl;
#endif
    return Etot;
}

//----------------------------------------------------------------------------//

void pot_water_ion::dip(const size_t& nw, const size_t& nI, double* dip)
{
    // requires that m_ttm4 has been previous called for this system
    m_ttm4.dip(nw, nI, dip);
}

//----------------------------------------------------------------------------//

void pot_water_ion::dipind(const size_t& nw, const size_t& nI, double* dipind)
{
    // requires that m_ttm4 has been previous called for this system
    m_ttm4.dipind(nw, nI, dipind);
}

//----------------------------------------------------------------------------//

} // namespace x2o

////////////////////////////////////////////////////////////////////////////////
