#include <cstdlib>

#include "system.h"

////////////////////////////////////////////////////////////////////////////////

namespace sys {

//------------------  Constructors / Destructors  ----------------------------//

// constructor
    system::system()
   {
   }

    // destructor
   system::~system()
    {
        // memory is only allocated if a molecule has been added to system
        if(molec.size() > 0){
            delete [] coords;
            delete [] charges;
            delete [] atmpolar;
            delete [] polfac;
        }
    }

    // copy constructor
   system::system(const system& o){
        for(size_t i = 0; i < o.molec.size(); ++i)
            system::add_molecule(o.molec[i]);

        system::create_system();

    }

//------------------  Tools for maintaining system  --------------------------//

    void system::add_molecule(std::shared_ptr<molecule> o){
        molec.push_back(o);
        system::create_system();
    }

    void system::set_pol(double* alpha){

    // First, copy alpha into molecules (sets both pol and polfac)
        int this_nsites(0);
        for(size_t n = 0; n < molec.size(); ++n){
        molec[n]->set_pol(alpha + this_nsites);
            this_nsites += molec[n]->get_nsites();
    }

    // Next, copy pol and polfac from molecules into the system
        for(size_t n = 0; n < molec.size(); ++n){
            double* molec_atmpolar = molec[n]->get_pol();
            double* molec_polfac   = molec[n]->get_polfac();

            const int shift = first_index_of_molec[n];

            for(size_t i = 0; i < molec[n]->get_nsites(); ++i){
                atmpolar[ i + shift]  = molec_atmpolar[i];
                polfac[   i + shift]  = molec_polfac[i];
            }
    }
    }

    void system::update_molecules(){
        int this_nsites(0);
        for(size_t n = 0; n < molec.size(); ++n){

            // copies positions from system into the molecule objects
            molec[n]->set_coords(coords + 3*this_nsites);

            // if the molecule has geometry dependent properties, 
            // copy these back into the system
            if(  molec[n]->has_virtual_sites()
              || molec[n]->has_charge_derivatives() ){

                // repositions auxiliary sites
                molec[n]->get_coords(coords + 3*this_nsites);
                // geometry dependent charges
                molec[n]->get_charges(charges + this_nsites);

            }
            this_nsites += molec[n]->get_nsites();
        }
    }

    void system::allocate(const int n){

        coords   = new double[3*n];
        charges  = new double[  n];
        atmpolar = new double[  n];
        polfac   = new double[  n];

        excluded12.clear();
        excluded13.clear();
        excluded14.clear();
        first_index_of_molec.clear();

    }

    void system::create_system(){
        if((molec.size() - 1) > 0){

            delete [] coords;
            delete [] charges;
            delete [] atmpolar;
            delete [] polfac;

        }

        nsites = 0;
        for(size_t n = 0; n < molec.size(); ++n){
            first_index_of_molec.push_back(nsites);
            nsites += molec[n]->get_nsites();
        }

        allocate(nsites);

        for(size_t n = 0; n < molec.size(); ++n){
            double* molec_coords   = molec[n]->get_coords();
            double* molec_charges  = molec[n]->get_charges();
            double* molec_atmpolar = molec[n]->get_pol();
            double* molec_polfac   = molec[n]->get_polfac();

            const int shift = first_index_of_molec[n];

            for(size_t i = 0; i < molec[n]->get_nsites(); ++i){
                for(size_t j = 0; j < 3; ++j){
                    coords[3*(i+shift) + j] = molec_coords[3*i + j];
                }
                charges[  i + shift]  = molec_charges[i];
                atmpolar[ i + shift]  = molec_atmpolar[i];
                polfac[   i + shift]  = molec_polfac[i];
            }

            excluded_set_type::iterator iter;

            for(iter  = molec[n]->excluded12.begin();
                iter != molec[n]->excluded12.end(); ++iter){
                size_t a = (*iter).first + shift;
                size_t b = (*iter).second + shift;

                excluded12.insert(std::make_pair(a, b));
            }

            for(iter  = molec[n]->excluded13.begin();
                iter != molec[n]->excluded13.end(); ++iter){
                size_t a = (*iter).first + shift;
                size_t b = (*iter).second + shift;

                excluded13.insert(std::make_pair(a, b));
            }
            
            for(iter  = molec[n]->excluded14.begin();
                iter != molec[n]->excluded14.end(); ++iter){
                size_t a = (*iter).first + shift;
                size_t b = (*iter).second + shift;

                excluded14.insert(std::make_pair(a, b));
            }

        }
    }

} // namespace sys
