#include "ttm4.h"
#include "ps.h"

////////////////////////////////////////////////////////////////////////////////

namespace {

const double gammaM = 0.426706882;
const double gamma1 = 1.0 - gammaM;
const double gamma2 = gammaM/2;
const double gamma21 = gamma2/gamma1;

inline void compute_M_site_crd
    (const double O[3], const double H1[3], const double H2[3], double M[3])
{
    for (size_t i = 0; i < 3; ++i)
        M[i] = gamma1*O[i] + gamma2*(H1[i] + H2[i]);
}

} // namespace

////////////////////////////////////////////////////////////////////////////////

namespace x2o {

//------------------  Constructor / Destructor / Initialization  -------------//

ttm4::ttm4() { // base class constructor is used
}

ttm4::~ttm4() { // memory is deleted in base class destructor
}

// Separate copy constructor is required for geometry dependent charges
ttm4::ttm4(const ttm4& o){

    // Allocate memory and copy values
    allocate(o.nsites);
    std::copy(o.memory, o.memory + 6*o.nsites + 3*9, memory);

    // Set pointers
    charge     = memory;
    coords     = memory + nsites;
    atmpolar   = memory + nsites + nsites*3;
    polfac     = memory + nsites + nsites*3 + nsites;
    charge_grd = memory + nsites + nsites*3 + nsites + nsites;
    definedCoords = definedCharges = definedPolarizability = true;

}

//----------------------------------------------------------------------------//

ttm4::ttm4(double* crd) {
    nsites = 4;
    allocate(nsites);

    set_coords(crd);

    set_charges(crd);

    double p[4] = {1.310, // polarO
                   0.294, // polarH
                   0.294, // polarH
                   0.0};  // polarM
    set_pol(p);

    excluded12.clear();
    excluded13.clear();

    excluded12.insert(std::make_pair(0, 1)); // O - H1
    excluded12.insert(std::make_pair(0, 2)); // O - H2
    excluded12.insert(std::make_pair(0, 3)); // O - M
    excluded13.insert(std::make_pair(1, 2)); // H1 - H2
    excluded12.insert(std::make_pair(1, 3)); // H1 - M
    excluded12.insert(std::make_pair(2, 3)); // H2 - M
}

//----------------------------------------------------------------------------//
//---  The following functions override the virtual fxns of the base class  --//
//----------------------------------------------------------------------------//

// Override base class allocate function to also store charge derivatives
void ttm4::allocate(const int n) {
    nsites = n;
    definedNsites = true;

    memory = new double [nsites     // charges
                        + nsites*3  // coords   
                        + nsites    // polarizabilities
                        + nsites    // polfacs
                        + 3*9];     // charge gradients (3 charges, 9 crds)
}

//----------------------------------------------------------------------------//

// Override base class set_coord fxn to add m-site
void ttm4::set_coords(double* new_crds) {
    coords = memory + nsites;

    // set O H H
    std::copy(new_crds, new_crds + 9, coords);

    // Calculate and set M-site (assumes ordering O H H)
    compute_M_site_crd(coords, coords + 3, coords + 6, coords + 9);

    definedCoords = true;
    virtual_sites = true;

    // Any time coords are set, recalculate geometry-dependent charges
    ttm4::set_charges(coords);
}

//----------------------------------------------------------------------------//

// Override base class set_charges to calculate geometry depented charges
void ttm4::set_charges(double* atmcrds) {
    charge = memory; 
    charge_grd = memory + 6*nsites; // NOTE: see allocate()

    double chgtmp[3];
    ps::dms_nasa(0.0, 0.0, 0.0, atmcrds, chgtmp, charge_grd, false);

    const double CHARGECON = constants::CHARGECON;

    charge[0] = 0.0;                                                     // O
    charge[1] = CHARGECON*(chgtmp[1] + gamma21*(chgtmp[1] + chgtmp[2])); // H1
    charge[2] = CHARGECON*(chgtmp[2] + gamma21*(chgtmp[1] + chgtmp[2])); // H2
    charge[3] = CHARGECON*(chgtmp[0]/(1.0 - gammaM));                    // M

    definedCharges = true;
    charge_derivatives = true;
}

//----------------------------------------------------------------------------//

void ttm4::set_pol(double* p){
    atmpolar = memory + nsites + nsites*3;
    polfac = memory + nsites + nsites*3 + nsites;

    for(size_t i = 0; i < nsites; ++i){
        atmpolar[i] = p[i];
    }

    polfac[0] = p[0]; // O
    polfac[1] = p[1]; // H
    polfac[2] = p[2]; // H
    polfac[3] = p[0]; // M (uses O polfac)

    definedPolarizability = true;
}

//----------------------------------------------------------------------------//

void ttm4::redistrib_frc_virtual_2_real(double* grad){

    // redistribute the M-site derivatives
    for (size_t k = 0; k < 3; ++k) {
        grad[0 + k] += gamma1*grad[9 + k]; // O
        grad[3 + k] += gamma2*grad[9 + k]; // H
        grad[6 + k] += gamma2*grad[9 + k]; // H
        grad[9 + k] = 0.0; // M
    }

}

//----------------------------------------------------------------------------//

void ttm4::charge_derivative_force(double* phi, double* grad){

    double grdq[27];
    std::fill(grdq, grdq + 27, 0.0);

#define DQ3(l,m,k) charge_grd[k + 3*(m + 3*l)]
#define GRDQ(l,m,k) grdq[k + 3*(m + 3*l)]

    for (size_t k = 0; k < 3; ++k) {
        GRDQ(0, 0, k) = DQ3(0, 0, k) + gamma21*(DQ3(0, 0, k) + DQ3(0, 1, k));
        GRDQ(1, 0, k) = DQ3(1, 0, k) + gamma21*(DQ3(1, 0, k) + DQ3(1, 1, k));
        GRDQ(2, 0, k) = DQ3(2, 0, k) + gamma21*(DQ3(2, 0, k) + DQ3(2, 1, k));

        GRDQ(0, 1, k) = DQ3(0, 1, k) + gamma21*(DQ3(0, 1, k) + DQ3(0, 0, k));
        GRDQ(1, 1, k) = DQ3(1, 1, k) + gamma21*(DQ3(1, 1, k) + DQ3(1, 0, k));
        GRDQ(2, 1, k) = DQ3(2, 1, k) + gamma21*(DQ3(2, 1, k) + DQ3(2, 0, k));

        GRDQ(0, 2, k) = DQ3(0, 2, k) - 2*gamma21*(DQ3(0, 0, k) + DQ3(0, 1, k));
        GRDQ(1, 2, k) = DQ3(1, 2, k) - 2*gamma21*(DQ3(1, 0, k) + DQ3(1, 1, k));
        GRDQ(2, 2, k) = DQ3(2, 2, k) - 2*gamma21*(DQ3(2, 0, k) + DQ3(2, 1, k));
    }

    for(size_t i = 0; i < 27; ++i)
        grdq[i] *= constants::CHARGECON;

    const size_t io  = 0;
    const size_t ih1 = 3;
    const size_t ih2 = 6;

    for (size_t k = 0; k < 3; ++k) {
        grad[ih1 + k] += GRDQ(0, 0, k)*phi[1]  // phi(h1)
                       + GRDQ(0, 1, k)*phi[2]  // phi(h2)
                       + GRDQ(0, 2, k)*phi[3]; // phi(M)

        grad[ih2 + k] += GRDQ(1, 0, k)*phi[1]  // phi(h1)
                       + GRDQ(1, 1, k)*phi[2]  // phi(h2)
                       + GRDQ(1, 2, k)*phi[3]; // phi(M)

        grad[ io + k] += GRDQ(2, 0, k)*phi[1]  // phi(h1)
                       + GRDQ(2, 1, k)*phi[2]  // phi(h2)
                       + GRDQ(2, 2, k)*phi[3]; // phi(M)
    }
}

//----------------------------------------------------------------------------//

} // namespace x2o
