#include <cmath>
#include <cassert>

#include <fstream>
#include <sstream>
#include <iomanip>
#include <iostream>
#include <stdexcept>
#include <vector>

#include <gsl/gsl_multimin.h>
#include "tang-toennies.h"
#include "wlsq.h"

#include "buckingham.h"
#include "dispersion.h"
#include "ttm4-es.h"
#include "ttm4.h"

#include "read-xyz.h"
#include "orient.h"

#include "molecule.h"
#include "system.h"
#include "electrostatics.h"

////////////////////////////////////////////////////////////////////////////////
namespace {

// Definition of xyz_frame in read_xyz.h
static std::vector<tools::xyz_frame> training_set;

// Vector with the dispersion energy
static std::vector<double> training_set_de;

// Vector with the electrostatic energy
 static std::vector<double> training_set_ee;

// Weights of the configurations
static double* ts_weights = 0;


//----------------------------------------------------------------------------//

int nparams = 0;
int num_nonlinear_params = 0;

double C6OX = 0;
double C6HX = 0;

////////////////////////////////////////////////////////////////////////////////

namespace linear {

//----------------------------------------------------------------------------//

static double* A(0);
static double* y(0);
static double* params(0);

//----------------------------------------------------------------------------//

// Memory allocation
void allocate() {
  A = new double[training_set.size()*nparams];
  y = new double[training_set.size()];

  params = new double[nparams];
}

//----------------------------------------------------------------------------//

// PRECONDITION: The function has as argument a gsl_vector with the
// parameters that must be optimized.
// POSTCONDITION: The function returns the sum of the square of the
// differences between the fitted and the original.
double compute_chisq_buckinham(const gsl_vector* X, void* unused) {

  // Obtaining the non lineal parameters to fit from the vector
  double BOX = X->data[0];
  double BHX = X->data[1];
//  double deltaOX = X->data[2];
//  double deltaHX = X->data[3];
  double deltaOX = X->data[0];
  double deltaHX = X->data[1];

  if (deltaOX < 0 || deltaHX < 0 ) {
    return 100000.0;
  }

  for (size_t n = 0; n < training_set.size(); ++n) {
    // Calculation of the distances Cation Anion
    const double rOX = tools::distance(training_set[n].xyz +   0,  // 
                     training_set[n].xyz + 9);
    const double rH1X = tools::distance(training_set[n].xyz +   3,  // 
                     training_set[n].xyz + 9);
    const double rH2X = tools::distance(training_set[n].xyz +   6,  // 
                     training_set[n].xyz + 9);

    double disp = x2o::dispersion(rOX,rH1X,rH2X,deltaOX,deltaHX,C6OX,C6HX);

    // Filling the vector y with the known energies: electrostatic
    // and the CCSD(T) one.
    y[n] = training_set[n].E - disp - training_set_ee[n];

    // Filling the matrix A
    for (size_t p = 0; p < (size_t)nparams; ++p){
      if (p == 0) {
        // CA term of the repulsion energy, positive
        A[p + n*nparams] = std::exp(-BOX*rOX);
      } else if (p == 1) {
        A[p + n*nparams] = std::exp(-BHX*rH1X) + std::exp(-BHX*rH2X);
      }
    }
  }

  // Declaring the variable that contains the sum of the square 
  // of the differences, and the rank
  double chisq;
  int rank;

  // Calling the lineal solver to solve for the lineal parameters
  kit::wlsq::solve(training_set.size(), nparams, A, y, ts_weights, 
                            params, chisq, rank);
  // Printing results
  std::cout << "<#> chisq = " << chisq
            << " : rank = " << rank
            << std::endl;

  // Cheking for negative parameters
  for (int i=0;i<nparams;i++) {
    if (params[i] < 0) {
      return 100000.0;
    }
  }


  // Calculating the new Chisq
  double errsum = 0;
  for (size_t k = 0; k < training_set.size(); ++k) {

    // Calculating distances
    double rOX = tools::distance(training_set[k].xyz +   0,  // 
                     training_set[k].xyz + 9);
    double rH1X = tools::distance(training_set[k].xyz +   3,  // 
                     training_set[k].xyz + 9);
    double rH2X = tools::distance(training_set[k].xyz +   6,  // 
                     training_set[k].xyz + 9);

    double dispersion_energy = x2o::dispersion(rOX,rH1X,rH2X,deltaOX,deltaHX,C6OX,C6HX);

    // Calculating the fitted Buckinham Potential
    double buckCA  = x2o::buckingham(BOX,params[0],rOX) + x2o::buckingham(BHX,params[1],rH1X) + x2o::buckingham(BHX,params[1],rH2X);
    double rep_energy = buckCA;

    // Calculating the interaction energy
    double int_energy = rep_energy + dispersion_energy + training_set_ee[k];

    errsum+= (std::pow(int_energy - training_set[k].E,2)) * ts_weights[k];
  }

  chisq = errsum;
  // Done!
  return chisq;

}


} // namespace linear

////////////////////////////////////////////////////////////////////////////////

} // namespace 

////////////////////////////////////////////////////////////////////////////////


int main(int argc, char** argv) {

  // Checking for the correct usage
  if (argc < 6) {
    std::cerr << "usage: fit-shortrange "
              << "ion_charge " // charge of the ion1
              << "ion_polarizability " // polarizability of the ion1
              << "ion_polfac " // polarizability factor of the ion1
              << "C6OX " // C6 for O-X
              << "C6HX " // C6 for H-X
              << "Energy_range " // Energy since the weight is 0
              << "x0[0] "
              << "x0[1] "
              << "configs1.xyz configs2.xyz ..."
              << std::endl;
    return 0;
  }

  // Reading input parameters
  C6OX = atof(argv[4]);
  C6HX = atof(argv[5]);

  // Reading the different xyz files from the command line
  // If xyz file not well formatted, the program exits with an error
  try {
    // Checking each one of the input files in the command line
    //FIXME Change i_start yo appropiate value
    for(int i = 9; i < argc; ++i){
      tools::read_xyz(argv[i], training_set);
      std::cerr << "Loaded " << argv[i] << ". Currently " << training_set.size()
                << " configurations" << std::endl;
    }
  } catch (const std::exception& e) {
    std::cerr << e.what() << std::endl;
  }
  
  
  
  // Calculating the electrostatic potentiali for all the configs 
  for (size_t k = 0; k < training_set.size(); ++k){

    // First of all we define the two molecules.
    double *coords1 = training_set[k].xyz + 0;
    double *coords2 = training_set[k].xyz + 9;          
    molecule ion1;
    ion1.allocate(1);
    ion1.set_coords(coords2);
    double ch=atof(argv[1]);
    ion1.set_charges(&ch);
    double pol=atof(argv[2]);
    ion1.set_pol(&pol);
    double polfac=atof(argv[3]);
    ion1.set_polfac(&polfac);

    // We add the two molecules into the system
    sys::system my_system;
    {
      std::shared_ptr<molecule> tmp = std::make_shared<x2o::ttm4>(coords1);
      tmp->id = "n";
      my_system.add_molecule(tmp);
    }
    {
      std::shared_ptr<molecule> tmp = std::make_shared<molecule>(ion1);
      tmp->id = "i";
      my_system.add_molecule(tmp);
    }
    
    

    // We calculate the electrostatics energy
    ttm::electrostatics elect;
    double *grad = 0;
    ttm::smear_ttm4x smr;
    double energy = elect(my_system.nsites, my_system.charges,
                            my_system.polfac, my_system.atmpolar,
                            my_system.coords,
                            my_system.excluded12, my_system.excluded13,
                            my_system.first_index_of_molec,
                            my_system.molec, smr, grad);
    training_set_ee.push_back(energy);
  }


  // Weights of the differenc configurations
  ts_weights = new double[training_set.size()];

  // Looking for the absolute minimum energy 
  double E_min = 10000.0;
  for (size_t p = 0; p < training_set.size(); ++p){
    if (E_min > training_set[p].E) {
      E_min = training_set[p].E;
    }
  }
  // Reading DE, setting the weights 
  double DE = atof(argv[6]);
  for (size_t p = 0; p < training_set.size(); ++p){
//    ts_weights[p] = 1.0 - pow(DE / ( training_set[p].E - E_min + DE ),2);
    ts_weights[p] = pow(DE / ( training_set[p].E - E_min + DE ),2);
  }
  
  nparams = 2; // Number of linear parameters
  num_nonlinear_params = 2; // Number of non linear parameters

  // Starting the preparation for the non linear fitting
  // Allocation of memory for:
  // A -> Matrix as a vector corresponding to the coef. of each non-linear
  // parameter (dimension: train_size x nparams)
  // y -> The values to fit (dimesion: train_size)
  linear::allocate();

  // Using the GSL library
  // This vector will contain the non linear parameters to fit
  // In order: 1. deltaCA/BCA (they are the same)
  gsl_vector* x = gsl_vector_alloc(num_nonlinear_params);

  x->data[0] = atof(argv[7]);
  x->data[1] = atof(argv[8]);
//  x->data[2] = atof(argv[7]);
//  x->data[3] = atof(argv[8]);

  // Definition of the function to minimize via multiminimization
  gsl_multimin_function chisq_func;

  // Defining the cpp function that want to minimize,
  // and the number of parameters to modify
  chisq_func.n = num_nonlinear_params;
  chisq_func.f = linear::compute_chisq_buckinham;

  // Creating a pointer to a vector of size num_nonlinear_params
  // This is the initial stepsize
  gsl_vector* ss = gsl_vector_alloc(num_nonlinear_params);
  
  // Setting the initial values of the vector at 0.1
//  gsl_vector_set_all(ss, 0.01); 
  ss->data[0] = atof(argv[7]);
  ss->data[1] = atof(argv[8]);
//  ss->data[2] = atof(argv[7]);
//  ss->data[3] = atof(argv[8]);

  // Starting simplex method
  // Printing the initial data
  std::cout << "\n<> initial simplex sides:\n";
  for (size_t n = 0; n < (size_t)num_nonlinear_params; ++n) {
    std::cout << n << " : " << ss->data[n] << "\n";
  }

  // This function returns a pointer to a newly allocated instance of 
  // a minimizer of type nmsimplex2rand for an num_nonlinear_params-dimension 
  // function.
  gsl_multimin_fminimizer* s =
        gsl_multimin_fminimizer_alloc(gsl_multimin_fminimizer_nmsimplex2rand,
        num_nonlinear_params);

  // This function initializes the minimizer s to minimize the function 
  // chisq_func, starting from the initial point x. The size of the 
  // initial trial steps is given in vector ss.
  gsl_multimin_fminimizer_set(s, &chisq_func, x, ss);

  // Initializing the number of iterations and the status
  int iter(0), status(0);

  // Solving the problem iteratively
  do {
    // Setting the iterations increment and checking the status
    // This function performs a single iteration of the minimizer s. 
    // If the iteration encounters an unexpected problem then an error 
    // code will be returned.
    ++iter;
    status = gsl_multimin_fminimizer_iterate(s);

    if (status) {
      break;
    }

    // This function tests the minimizer specific characteristic size 
    // (if applicable to the used minimizer) against absolute tolerance 
    // epsabs. The test returns GSL_SUCCESS if the size is smaller 
    // than tolerance, otherwise GSL_CONTINUE is returned.
//    const double size = linear::compute_chisq_buckinham(s->x, 0);
    const double size = gsl_multimin_fminimizer_size(s);
    status = gsl_multimin_test_size(size, 1e-5);
//    status = gsl_multimin_test_size(size, 1.0);

    // If the tolerance is small enought, we are done
    if (status == GSL_SUCCESS) {
      std::cout << "!!! converged !!!" << std::endl;
    }

    // Printing intermediate results
    if (iter > 0 && iter%10 == 0) {
      std::cout << "\n<> solution:\n";
      for (size_t n = 0; n <  (size_t)num_nonlinear_params; ++n) {
        std::cout << n << " : " << s->x->data[n] << "\n";
      }
      std::cout << "<>" << std::endl;
    }
    
    //Check if next iteration is necessary!
  } while (status == GSL_CONTINUE && iter < 500);

  // Declaration of the final adjusted parameters
  const double final_chisq = linear::compute_chisq_buckinham(s->x, 0);
  const double AOX = linear::params[0];
  const double AHX = linear::params[1];
  const double BOX = s->x->data[0];
  const double BHX = s->x->data[1];
//  const double deltaOX = s->x->data[2];
//  const double deltaHX = s->x->data[3];
  const double deltaOX = s->x->data[0];
  const double deltaHX = s->x->data[1];

  // Printing the final parameters
  std::cout << "Nonlinear parameters: "
        << "BOX = " << BOX << ' '
        << "BHX = " << BHX << ' '
        << "deltaOX = " << deltaOX << ' '
        << "deltaHX = " << deltaHX << std::endl;
  std::cout << "Linear parameters: "
        << "AOX = " << AOX << ' '
        << "C6OX = " << C6OX << ' '
        << "AHX = " << AHX << ' '
        << "C6HX = " << C6HX  << ' '
        << "Polarizability: " << atof(argv[2]) << std::endl;
  std::cout << "Error per configurat: "
            << final_chisq/(training_set.size()) << std::endl;


  // Recalculating the curve again, and printing the different energies
  std::cout.precision(8); // Setting precission 8 decimals
  // Setting the headers
  std::cout << std::setw(18) << std::right << "# O-X dist."
            << std::setw(18) << std::right << "# M:Rep"
            << std::setw(18) << std::right << "# R:Rep"
            << std::setw(18) << std::right << "# M:Rep + Disp"
            << std::setw(18) << std::right << "# R:Rep + Disp"
            << std::setw(18) << std::right << "# R:Elec_energ"
            << std::setw(18) << std::right << "# M:Int_energ"
            << std::setw(18) << std::right << "R:Int_energ"  << std::endl;

  double errsum = 0;
  for (size_t k = 0; k < training_set.size(); ++k) {

    // Calculating distances
    double rOX = tools::distance(training_set[k].xyz +   0,  // 
                     training_set[k].xyz + 9);
    double rH1X = tools::distance(training_set[k].xyz +   3,  // 
                     training_set[k].xyz + 9);
    double rH2X = tools::distance(training_set[k].xyz +   6,  // 
                     training_set[k].xyz + 9);

    double dispersion_energy = x2o::dispersion(rOX,rH1X,rH2X,deltaOX,deltaHX,C6OX,C6HX);

    // Calculating the fitted Buckinham Potential
    double buckCA  = x2o::buckingham(BOX,AOX,rOX) + x2o::buckingham(BHX,AHX,rH1X) + x2o::buckingham(BHX,AHX,rH2X);
    double rep_energy = buckCA;

    // Calculating the interaction energy
    double int_energy = rep_energy + dispersion_energy + training_set_ee[k];

    // Printing all the energies
    std::cout << std::fixed
              << std::setw(18) << std::right << rOX
              << std::setw(18) << std::right << rep_energy
              << std::setw(18) << std::right << training_set[k].E - training_set_ee[k] - dispersion_energy
              << std::setw(18) << std::right << rep_energy + dispersion_energy
              << std::setw(18) << std::right << training_set[k].E - training_set_ee[k]
              << std::setw(18) << std::right << training_set_ee[k]
              << std::setw(18) << std::right << int_energy
              << std::setw(18) << std::right << training_set[k].E << std::endl;

    errsum+= (std::pow(int_energy - training_set[k].E,2)) * ts_weights[k];
  }
  std::cout << "manual chisq: " << errsum << std::endl;  

  // Check if all the fitted parameters are positive:
  if (AOX < 0 || C6OX < 0 || BOX < 0 || AHX < 0 || C6HX < 0 || BHX < 0 ) {
    std::cerr << "WARNING: SOME OF THE LINEAR PARAMETERS "
              << "ARE NEGATIVE." << std::endl;
    std::cerr << "EXITING NOW... ... ..." << std::endl;
    return 1;
  }


  return 0;
}
















