#include <cmath>
#include <cassert>

#include <fstream>
#include <sstream>
#include <iomanip>
#include <iostream>
#include <stdexcept>
#include <vector>

#include <gsl/gsl_multimin.h>
#include "tang-toennies.h"

#include "buckingham.h"
#include "dispersion.h"
#include "ttm4-es.h"
#include "ttm4.h"

#include "read-xyz.h"
#include "orient.h"

#include "molecule.h"
#include "system.h"
#include "electrostatics.h"

////////////////////////////////////////////////////////////////////////////////
namespace {

// Vector with the dispersion energy
static std::vector<double> training_set_de;

// Vector with the electrostatic energy
 static std::vector<double> training_set_ee;

////////////////////////////////////////////////////////////////////////////////

} // namespace 

////////////////////////////////////////////////////////////////////////////////


int main(int argc, char** argv) {

  // Checking for the correct usage
  if (argc < 11) {
    std::cerr << "usage: fit-shortrange "
              << "ion_charge " // charge of the ion1
              << "ion_polarizability " // polarizability of the ion1
              << "ion_polfac " // polarizability factor of the ion1
              << "AOX_buck " // A parameter
              << "AHX_buck " // A parameter
              << "C6OX " // C6
              << "C6HX " // C6
              << "deltaOX " // They are the same
              << "deltaHX " // They are the same
              << "theta(deg) " // They are the same
              << "phi(deg) " // They are the same
              << std::endl;
    return 0;
  }

  // Reading input parameters
  double AOX = atof(argv[4]);
  double AHX = atof(argv[5]);
  double C6OX = atof(argv[6]);
  double C6HX = atof(argv[7]);
  double BOX = atof(argv[8]);
  double BHX = atof(argv[9]);
  double theta = atof(argv[10]);
  double phi = atof(argv[11]);

  double pi = 3.14159265359;

  size_t npts = 200;
  double xyz[3*npts];
  
  double water_coord[] = {0.0,         0.0,        0.0, 
                          0.59357450, -0.76924354, 0.0,
                          0.59357450,  0.76924354, 0.0};

  double th = theta/360*2*pi;
  double ph = phi/360*2*pi;

  for (size_t k=0; k < npts; k++) {
    double r = (double)k/10.0+1;
//    std::cout << d << std::endl;
    xyz[3*k + 0] = r * sin(th) * cos(ph);
    xyz[3*k + 1] = r * sin(th) * sin(ph);
    xyz[3*k + 2] = r * cos(th);
  }
  
  double  test_pol = 0;
  double  test_polfac = 0;
  
//  std::cout << xyz[3*10 + 5] << std::endl;
  // Calculating the electrostatic potentiali for all the configs 
  for (size_t k = 0; k < npts; ++k){

    // First of all we define the two molecules.
    double *coords = xyz + 3*k;
    molecule ion;
    ion.allocate(1);
    ion.set_coords(coords);
    double ch=atof(argv[1]);
    ion.set_charges(&ch);
    double pol=atof(argv[2]);
    ion.set_pol(&pol);
    double polfac=atof(argv[3]);
    ion.set_polfac(&polfac);

    // We add the two molecules into the system
    sys::system my_system;
    {
      std::shared_ptr<molecule> tmp = std::make_shared<x2o::ttm4>(water_coord);
      tmp->id = "n";
      my_system.add_molecule(tmp);
    }

    {
       std::shared_ptr<molecule> tmp = std::make_shared<molecule>(ion);
       tmp->id = "i";
       my_system.add_molecule(tmp);
    }
    

    ion.get_pol(&test_pol);
    ion.get_polfac(&test_polfac);


    // We calculate the electrostatics energy
    ttm::electrostatics elect;
    double *grad = 0;
    ttm::smear_ttm4x smr;
    double energy = elect(my_system.nsites, my_system.charges,
                            my_system.polfac, my_system.atmpolar,
                            my_system.coords,
                            my_system.excluded12, my_system.excluded13,
                            my_system.first_index_of_molec,
                            my_system.molec, smr, grad);
    training_set_ee.push_back(energy);
  }

  const double deltaOX = BOX;
  const double deltaHX = BHX;

  // Recalculating the curve again, and printing the different energies
  std::cout.precision(8); // Setting precission 8 decimals
  std::cout << "Parameters: AOX= " << AOX 
            << " AHX= " << AHX
            << " C6OX= " << C6OX
            << " C6HX= " << C6HX << std::endl
            << " BOX= " << BOX
            << " BHX= " << BHX
            << " deltaOX= " << deltaOX
            << " deltaHX= " << deltaHX << std::endl
            << " pol= " << test_pol
            << " polfac= " << test_polfac << std::endl;
  // Setting the headers
  std::cout << std::setw(18) << std::right << "# X-O dist."
            << std::setw(18) << std::right << "# Rep"
            << std::setw(18) << std::right << "# Disp"
            << std::setw(18) << std::right << "# Elec_energ"
            << std::setw(18) << std::right << "Int_energ"  << std::endl;

  for (size_t k = 0; k < npts; ++k) {

    // Calculating distances
    double   rOX  = tools::distance(water_coord + 0 , xyz + 3*k );
    double   rH1X = tools::distance(water_coord + 3 , xyz + 3*k );
    double   rH2X = tools::distance(water_coord + 6 , xyz + 3*k );

    // Calculating the fitted Buckinham Potential
    double rep_energy  = x2o::buckingham(BOX,AOX,rOX) 
                         + x2o::buckingham(BHX,AHX,rH1X) 
                         + x2o::buckingham(BHX,AHX,rH2X);

    // Calculating the dispersion energy with the fitted parameters
    double dispersion_energy = x2o::dispersion(rOX, deltaOX, C6OX)
                               + x2o::dispersion(rH1X, deltaHX, C6HX) 
                               + x2o::dispersion(rH2X, deltaHX, C6HX);

    // Calculating the interaction energy
    double int_energy = rep_energy + dispersion_energy + training_set_ee[k];

    // Printing all the energies
    std::cout << std::fixed
              << std::setw(18) << std::right << rOX
              << std::setw(18) << std::right << rep_energy
              << std::setw(18) << std::right << dispersion_energy
              << std::setw(18) << std::right << training_set_ee[k]
              << std::setw(18) << std::right << int_energy
              << std::endl;

  }

  return 0;
}
















